# ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::: #
# Date: 25.09.2016                                                            #
# Author: Ole-Johan Skrede                                                    #
#                                                                             #
# Solution proposal as part of the exercise program in                        #
# INF4300 - Digital image analysis at the University of Oslo                  #
#                                                                             #
# ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::: #
"""
Solution proposal for exercise 5, task 3.
"""

# pylint: disable=expression-not-assigned
# pylint: disable=bad-indentation
# pylint: disable=redefined-outer-name
# pylint: disable=import-error
# pylint: disable=singleton-comparison

import os

import numpy as np
import matplotlib.pyplot as plt
import cv2

import chain_code # Using the method from week 4. Remember to move
                  # chain_code.py to this directory, or make a symlink.
import bit_quad_pattern_matching as bqpm

def perimeter_from_chain_code(cc, method='Freeman'):
  """Return perimeter"""
  num_odd = 0
  num_even = 0
  num_corners = 0 # Number of occurances of unequal consecutive chain elements

  for list_el in cc:
    num_odd += np.mod(list_el, 2)
  num_even = len(cc) - num_odd

  for i, _ in enumerate(cc):
    if i > 0:
      if not cc[i] == cc[i-1]:
        num_corners += 1

  if method == 'Freeman':
    return num_even + np.sqrt(2)*num_odd
  if method == 'Vossepoel':
    return 0.980*num_even + 1.406*num_odd - 0.091*num_corners
  if method == 'Kulpa':
    return np.pi/8 * (1 + np.sqrt(2)) * (num_even + np.sqrt(2)*num_odd)

def area_from_chain_code(cc):
  """Return area"""
  area = 0
  y_im1 = 0 # Arbitrary start point
  for _, c in enumerate(cc):
    c_ix, c_iy = chain_code.dir_to_coord(c, connectivity=8)
    area += c_ix*(y_im1 + c_iy/2)
    y_im1 += c_iy
  return area

def props_from_pattern_matching(image, method='Gray'):
  """Return area and perimeter computed by pattern matching"""

  num_matches = bqpm.count_matches(image, solid=False, seed_point=(24, 24),
                                   interior=False)

  if method == '4way':
    perimeter = 2*num_matches['hor'] + 2*num_matches['ver']
    area = num_matches['pixel']
  if method == 'Gray':
    perimeter = (num_matches['Q1'] + num_matches['Q2'] + num_matches['Q3'] +
                 2*num_matches['QD'])
    area = (num_matches['Q1'] + 2*num_matches['Q2'] + 3*num_matches['Q3'] +
            4*num_matches['Q4'] + 2*num_matches['QD']) / 4
  elif method == 'Duda':
    perimeter = (num_matches['Q2'] + (num_matches['Q1'] + num_matches['Q3'] +
                                      2*num_matches['QD'])/np.sqrt(2))
    area = (1/4*num_matches['Q1'] + 1/2*num_matches['Q2'] + 7/8*num_matches['Q3'] +
            num_matches['Q4'] + 3/4*num_matches['QD'])


  return area, perimeter

def area_from_radius(radius):
  """Compute area from radius"""
  return np.pi*radius*radius

def perimeter_from_radius(radius):
  """Compute perimeter from radius"""
  return 2*np.pi*radius


def main():
  """Connected components"""

  print('='*80)
  print('Solution to weekly exercises in INF4300')
  print('Exercise 5')
  print('Task 3')
  print('-'*80)

  # Configurations
  #result_dir = './results/task01'
  image_dir = '../../images/'
  #write_results = True

  image_names = ['circle2.png', 'circle5.png', 'circle15.png']
  filenames = [os.path.join(image_dir, name) for name in image_names]

  perimeters = {}
  perimeters['contour_naive'] = []
  perimeters['contour_opencv'] = []
  perimeters['cc_freeman'] = []
  perimeters['cc_vossepoel'] = []
  perimeters['cc_kulpa'] = []
  perimeters['pm_4way'] = []
  perimeters['pm_gray'] = []
  perimeters['pm_duda'] = []

  areas = {}
  areas['contour_opencv'] = []
  areas['cc_freeman'] = []
  areas['pm_4way'] = []
  areas['pm_gray'] = []
  areas['pm_duda'] = []

  for filename in filenames:
    img = cv2.imread(filename, cv2.IMREAD_GRAYSCALE)

    # Properties from contour
    _, contours, _ = cv2.findContours(img, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_NONE)
    contour = contours[0]

    delta_squared = np.diff(contour, axis=0)*np.diff(contour, axis=0)
    perimeters['contour_naive'].append(np.sum(np.sqrt(np.sum(delta_squared, axis=1))))
    perimeters['contour_opencv'].append(cv2.arcLength(contour, True))

    areas['contour_opencv'].append(cv2.contourArea(contour))

    # Properties from chain code
    # Use own implementation from week 4
    cc, _ = chain_code.trace_boundary(img, connectivity=8, background=0)

    perimeters['cc_freeman'].append(perimeter_from_chain_code(cc, 'Freeman'))
    perimeters['cc_vossepoel'].append(perimeter_from_chain_code(cc, 'Vossepoel'))
    perimeters['cc_kulpa'].append(perimeter_from_chain_code(cc, 'Kulpa'))

    areas['cc_freeman'].append(area_from_chain_code(cc))

    # Properties from bit quad pattern matching
    area, perim = props_from_pattern_matching(img, method='4way')
    areas['pm_4way'].append(area)
    perimeters['pm_4way'].append(perim)

    area, perim = props_from_pattern_matching(img, method='Gray')
    areas['pm_gray'].append(area)
    perimeters['pm_gray'].append(perim)

    area, perim = props_from_pattern_matching(img, method='Duda')
    areas['pm_duda'].append(area)
    perimeters['pm_duda'].append(perim)

  # Plot images
  circle2 = cv2.imread(filenames[0], cv2.IMREAD_GRAYSCALE)
  circle5 = cv2.imread(filenames[1], cv2.IMREAD_GRAYSCALE)
  circle15 = cv2.imread(filenames[2], cv2.IMREAD_GRAYSCALE)

  fig, ax1 = plt.subplots()
  #ax1 = fig1.add_subplot(1, 1, 1)
  ax1.imshow(circle2, cmap='gray', interpolation='none')
  ax1.set_xticks(np.arange(0, circle2.shape[1]))
  ax1.set_xticks(np.arange(0.5, circle2.shape[1] + 0.5), minor=True)
  ax1.get_xaxis().set_ticks([]) # Hide tick labels
  ax1.set_yticks(np.arange(0, circle2.shape[0]))
  ax1.set_yticks(np.arange(0.5, circle2.shape[0] + 0.5), minor=True)
  ax1.get_yaxis().set_ticks([]) # Hide tick labels
  ax1.grid(b=True, which='minor', linestyle='-', color='gray')
  plt.tight_layout()
  fig.savefig('results/task03/circle2_grid.png', bbox_inches='tight', pad_inches=0)

  fig, ax2 = plt.subplots()
  #ax2 = fig.add_subplot(1, 3, 2)
  ax2.imshow(circle5, cmap='gray', interpolation='none')
  ax2.set_xticks(np.arange(0, circle5.shape[1]))
  ax2.set_xticks(np.arange(0.5, circle5.shape[1] + 0.5), minor=True)
  ax2.get_xaxis().set_ticks([]) # Hide tick labels
  ax2.set_yticks(np.arange(0, circle5.shape[0]))
  ax2.set_yticks(np.arange(0.5, circle5.shape[0] + 0.5), minor=True)
  ax2.get_yaxis().set_ticks([]) # Hide tick labels
  ax2.grid(b=True, which='minor', linestyle='-', color='gray')
  plt.tight_layout()
  fig.savefig('results/task03/circle5_grid.png', bbox_inches='tight', pad_inches=0)

  fig, ax3 = plt.subplots()
  #ax3 = fig.add_subplot(1, 3, 3)
  ax3.imshow(circle15, cmap='gray', interpolation='none')
  ax3.set_xticks(np.arange(0, circle15.shape[1]))
  ax3.set_xticks(np.arange(0.5, circle15.shape[1] + 0.5), minor=True)
  ax3.get_xaxis().set_ticks([]) # Hide tick labels
  ax3.set_yticks(np.arange(0, circle15.shape[0]))
  ax3.set_yticks(np.arange(0.5, circle15.shape[0] + 0.5), minor=True)
  ax3.get_yaxis().set_ticks([]) # Hide tick labels
  ax3.grid(b=True, which='minor', linestyle='-', color='gray')
  plt.tight_layout()
  fig.savefig('results/task03/circle15_grid.png', bbox_inches='tight', pad_inches=0)


  # Comparison of methods

  print('Perimeter')
  print('{0:<14} {1:>10} {2:>10} {3:>10}'.format('Method',
                                                 os.path.splitext(image_names[0])[0],
                                                 os.path.splitext(image_names[1])[0],
                                                 os.path.splitext(image_names[2])[0]))
  print('{0:<14} {1:>10,.2f} {2:>10,.2f} {3:>10,.2f}'.format('From radius',
                                                             perimeter_from_radius(2),
                                                             perimeter_from_radius(5),
                                                             perimeter_from_radius(15)))
  for m, p in sorted(perimeters.items()):
    print('{0:<14} {1:>10,.2f} {2:>10,.2f} {3:>10,.2f}'.format(m, p[0], p[1], p[2]))

  print('Area')
  print('{0:<14} {1:>10} {2:>10} {3:>10}'.format('Method',
                                                 os.path.splitext(image_names[0])[0],
                                                 os.path.splitext(image_names[1])[0],
                                                 os.path.splitext(image_names[2])[0]))
  print('{0:<14} {1:>10,.2f} {2:>10,.2f} {3:>10,.2f}'.format('From radius',
                                                             area_from_radius(2),
                                                             area_from_radius(5),
                                                             area_from_radius(15)))
  for m, a in sorted(areas.items()):
    print('{0:<14} {1:>10,.2f} {2:>10,.2f} {3:>10,.2f}'.format(m, a[0], a[1], a[2]))



  plt.show()

if __name__ == "__main__":
  main()
